from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import silhouette_score
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.patches import Patch
from matplotlib import pyplot as plt

def genemodules(self, nmax=6, use_grouped=False, n_modules = None):
    if use_grouped:
        df1 = self.co_expression_grouped.copy()
    else:
        df1 = self.co_expression.copy()

    df2 = df1.copy()
    df2.rename(columns={'gene_1': 'gene_2', 'gene_2': 'gene_1'}, inplace=True)

    coexpr = pd.concat([df1, df2])
    coexpr = coexpr.pivot(index='gene_1', columns='gene_2', values='L')
    coexpr.index.name=None
    coexpr.columns.name=None
    np.fill_diagonal(coexpr.values, 0)
    X = coexpr.values.copy()

    if n_modules is None:
        range_n_clusters = list(range(2, nmax+1))
        sil_scores = []

        for n_clusters in range_n_clusters:
            hc = AgglomerativeClustering(n_clusters=n_clusters)
            cluster_labels = hc.fit_predict(X)
            silhouette_avg = silhouette_score(X, cluster_labels)
            sil_scores.append(silhouette_avg)
            print(
                "For n_clusters =",
                n_clusters,
                "The average silhouette_score is :",
                silhouette_avg,
            )
        sil_scores_r = sil_scores[::-1]
        i = len(sil_scores_r) - np.argmax(sil_scores_r) - 1
        nclust_opt = range_n_clusters[i]
    else:
        nclust_opt = n_modules
    
    print("Number of selected clusters = ", nclust_opt)

    hc_opt = AgglomerativeClustering(n_clusters=nclust_opt)
    cluster_labels_opt = hc_opt.fit_predict(X)
    silhouette = silhouette_score(X, cluster_labels_opt)
    cluster_labels_opt = [f'm{c}' for c in cluster_labels_opt]
    module2Genes = {}
    gene2module = {}

    for gene, lab in zip(coexpr.index.tolist(), np.ravel(cluster_labels_opt).tolist()):
        if lab not in module2Genes.keys():
            module2Genes[lab] = [gene]
        else:
            module2Genes[lab].append(gene)
        
        gene2module[gene] = lab

    module_df = pd.DataFrame({"label": cluster_labels_opt, "silhouette": silhouette}, 
                                    index=coexpr.index.copy())
    sg.gene_modules = module_df
    
    colors = {'m0': "green", 'm1': "magenta", 'm2': "darkorange",
                 'm3': "gray", 'm4': "beige", 'm5': "pink",
                  'm6': "darkgray", 'm7': "yellow",
                  'm8' : "red", 'm9': "blue"}

    g = sns.clustermap(coexpr, figsize=(6,6), cmap='bwr',
                vmin=-0.8, vmax=0.8,yticklabels=1, xticklabels=1,
                row_colors=[colors[l] for l in cluster_labels_opt])
    
    uniq_modules = list(set(cluster_labels_opt))
    handles = [Patch(facecolor=colors[l]) for l in uniq_modules]
    plt.legend(handles, uniq_modules, title='Gene Modules',
            bbox_to_anchor=(1.2, 1), bbox_transform=plt.gcf().transFigure,
             loc='upper right')

    mod_dict = {'gene2mod': gene2module, 'mod2gene': module2Genes, 'module_df': module_df, 'module_clustmap': g}
    self.module_dict = mod_dict


# import scanpy as sc
# import anndata as ad
# import squidpy as sq
# import pandas as pd
# import seaborn as sns
# import numpy as np
import warnings
from matplotlib import pyplot as plt
warnings.filterwarnings("ignore")
from ._utils import *
from ._gmod import *
from ._plotting import *

class SG(object):
    """ Spatial Enrichment Analysis of Gene Association using L-index.
    """
    def __init__(self, 
                 count_path = "", meta_path = "",
                 visium_path="/mnt/atlas_local/linhua/data/SEAGAL/test_data/"):
        


        data = load_raw(self, count_path, meta_path, visium_path)
            
        ## preprocess, and QC
        data = process_st(self, data, min_counts=150, min_cells=10)
        ## spatial process
        self.adata = spatial_process(data)
        self.co_expression = None
        self.co_expression_grouped = None
        self.hotspot = None
        self.hotspot_names = None
        # self.plot =  plot()









from spatial_measure import Local_L
import scanpy as sc
from matplotlib import pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd

def hotspot(self, var_1, var_2, dropout_rm=True, use_grouped=False, cmap='viridis', alpha_img=0.5):
    """ Calculate local L-index for var pairs (var_1 and var_2). 
    """

    if use_grouped:
        adata = self.grouped_adata.copy()
    else:
        adata = self.adata.copy()
    # calculate local L-index for specific var_name in adata
    ### 
    try:
        adata.X = adata.X.toarray()
    except:
        pass

    if var_1 not in adata.var_names or var_2 not in adata.var_names:
        print("One of the queried features is not in the adata. Check if the correct `use_grouped` parameter is used, or typos in the names.")
        return

    adata1 = Local_L(adata, var_1, var_2, dropout_rm=dropout_rm, max_RAM=32)
    adata.obs[f'{var_1} & {var_2}'] = adata1.uns['Local_L'].ravel()
    f, ax = plt.subplots(1,1, figsize=(4,4))
    sc.pl.spatial(adata, color=f'{var_1} & {var_2}', ax=ax, cmap=cmap, alpha_img=alpha_img)
    if use_grouped:
        self.grouped_adata = adata.copy()
    else:
        self.adata = adata.copy()
    return f, adata.obs[f'{var_1} & {var_2}']

def volcano(self, FDR_cutoff=0.05, L_cutoff=0.1, use_grouped=False):
    if use_grouped:
        df = self.co_expression_grouped.copy()
    else:
        df = self.co_expression.copy()
    
    vals  = df['L.FDR'].copy()
    vals = vals[vals > 0]

    if len(vals) > 0:
        nonzeromin = min(vals)
        df['-log10(FDR)'] = -np.log10(df['L.FDR'] + nonzeromin)  
    else:
        df['-log10(FDR)'] = -2
        
    df["Association"] = "NS"
    df.loc[(df['L.FDR'] <= FDR_cutoff) & (df['L'] >= L_cutoff), "Association"] = "SigPos"
    df.loc[(df['L.FDR'] <= FDR_cutoff) & (df['L'] <= -L_cutoff), "Association"] = "SigNeg"

    f, ax = plt.subplots(1,1, figsize=(4,3))
    sns.scatterplot(data=df, x='L', y='-log10(FDR)', 
                    hue='Association',
                        palette={'NS': 'lightgray', 'SigPos': 'red', 'SigNeg': 'blue'},
                        s=20, ax=ax)
    sns.despine(trim=True, ax=ax)
    return f

def clustermap(self, figsize=(6,6), use_grouped=False):
    if use_grouped:
        df1 = self.co_expression_grouped.copy()
    else:
        df1 = self.co_expression.copy()

    df2 = df1.copy()
    df2.rename(columns={'gene_1': 'gene_2', 'gene_2': 'gene_1'}, inplace=True)

    coexpr = pd.concat([df1, df2])
    coexpr = coexpr.pivot(index='gene_1', columns='gene_2', values='L')
    coexpr.index.name=None
    coexpr.columns.name=None
    np.fill_diagonal(coexpr.values, 0)
    if not use_grouped:
        sns.clustermap(coexpr, figsize=figsize, cmap='bwr', annot=False,
               vmin=-0.8, vmax=0.8)
    else:
        annot = coexpr.round(2).astype(str)
        for (gene1, gene2, p) in zip(df1.gene_1.tolist(), df1.gene_2.tolist(), df1['L.FDR'].tolist()):
            if p <= 0.01:
                if "*" not in annot.loc[gene1, gene2]:
                    annot.loc[gene1, gene2] = f"{annot.loc[gene1, gene2]}**"
                if "*" not in annot.loc[gene2, gene1]:
                    annot.loc[gene2, gene1] = f"{annot.loc[gene2, gene1]}**"
            elif p <= 0.05:
                if "*" not in annot.loc[gene1, gene2]:
                    annot.loc[gene1, gene2] = f"{annot.loc[gene1, gene2]}*"
                if "*" not in annot.loc[gene2, gene1]:
                    annot.loc[gene2, gene1] = f"{annot.loc[gene2, gene1]}*"
            else:
                pass
        sns.clustermap(coexpr, figsize=figsize, cmap='bwr', 
                annot=annot, fmt='s',
               vmin=-0.8, vmax=0.8)
    return